#lang racket/base

(require "lascheme.rkt")

(provide (except-out (all-from-out "lascheme.rkt")
                     raise
                     error                     
                     with-handlers)
         (rename-out [levels/raise raise]
                     [levels/error error]
                     [levels/with-handlers with-handlers]))

;; Puts the current level as a continuation mark
;; that is captured by raise when raising exn values
(define (levels/raise v . barrier)
  (w-c-m 'current-level (current-level)
         (raise v barrier)))

;; Redefines error, putting the current level as a continuation mark
(define levels/error
  (lambda args
    (w-c-m 'current-level (current-level)
           (apply error args))))

;; exn -> int
;; Gets the execution level of the exception
;; Only exceptions capture the continuation mark with the current level.
(define (exception-level exn)   
  (car (continuation-mark-set->list
        (exn-continuation-marks exn) 
        'current-level)))

;; Inspects the level of the exception by observing
;; the continuation marks stored in the exn value.
;; Because of that, it works only with exceptions, not with other values raised.
(define-syntax (levels/with-handlers stx)
  (syntax-case stx (with-handlers)  
    ((levels/with-handlers ((predicate handler) ...) body)
     (with-syntax ([(handler-closure ...) (generate-temporaries #'(handler ...))])
       #'(let ((handler-closure handler) ...) ;; we evaluate only once to get the handler closure
           (with-handlers ([(lambda (exn) #t)
                            (lambda (exn)
                              (with-handlers
                                  ([(lambda (x)
                                      (and (predicate x)
                                           (= (procedure-level handler-closure)
                                              (exception-level exn)))) 
                                    handler-closure] ...                                                      [(lambda (x) #t) (lambda (x) (levels/raise exn))])
                                (raise exn)))])
             body))))))



