#lang racket


(require "exceptions.rkt")

;; test-exn doesn't work with levels
(require "test.rkt")

(define (level n) (lambda (l) (eq? l n)))
(define test-handler (λ (x)
                       (display x)
                       (newline)                       
                       #t))
(define test-handler-s (lambda-s (x) (display x) (newline) #t))

;; Catch exception at same level
(test (with-handlers ([(λ (x) #t)                         
                       (λ (x) 1)])
        (error 'a))
      1)

;; Catch exception at same level, shifted by down.
(test (^down (with-handlers ([(λ (x) #t) (λ (x) 2)])
               (error 'a)))
      2)

;; Catch exception at same level, shifted by up.
(test (^up (with-handlers ([(λ (x) #t) test-handler])
             (error 'a)))
      #t)

;;; Catch exception at same level, using level-capturing functions.
(test (with-handlers ([(λ (x) #t) (^down (lambda-s (x) (display x) (newline) #t))])
        (^down (error 'a)))
      #t)

(test (with-handlers ([(λ (x) #t) (^up (lambda-s (x) (display x) (newline) #t))])
        (^up (error 'a)))
      #t)

;; Raising and reraising
(test (with-handlers ([(λ (x) #t) test-handler])
        (^down (^up (error 'a))))
      #t)

(test (with-handlers ([(λ (x) #t) test-handler])
        (^up (^down (error 'a))))
      #t)

(test (with-handlers ([(λ (x) #t) test-handler])
        (^up (^down (error 'a))))
      #t)

(test (with-handlers ([(λ (x) #t) test-handler])
        (error (error 'a)))
      #t)

(test (with-handlers ([(λ (x) #t) test-handler])
        (error (error (error (error (error 'a))))))
      #t)

(test (with-handlers ([(λ (x) #t) (^up (lambda-s (x) (display x) (newline) #t))])
        (raise (^up (error 'a))))
      #t)

(test (with-handlers ([(λ (x) #t) (^down (lambda-s (x) (display x) (newline) #t))])
        (raise (^down (error 'a))))
      #t)

;;
;; Aspects Tests
;;
(define (test-function x) (error x))

;;; Aspect handler catches base exception
(let ((handler 
       (lambda (proceed)   ;; proceed
         (lambda ctx       ;; ctx
           (lambda (a)     ;; val
             (begin
               (printf "Advice. a: ~s - ctx: ~s - proceed: ~s" a ctx proceed)
               (newline)
               (with-handlers ([(λ(x) #t) (^down (lambda-s (x) (display x) (newline) #t))])
                 (proceed a))))))))  
  
  (test
   (deploy-fluid (aspect (call test-function) ;; pc
                         handler)
                 (test-function 'b))
   #t))

;; Other way of doing the same, without level-capturing functions
(let ((handler 
       (lambda (proceed)   ;; proceed
         (lambda ctx       ;; ctx
           (lambda (a)     ;; val
             (begin
               (printf "Advice. a: ~s - ctx: ~s - proceed: ~s" a ctx proceed)
               (newline)
               (^down (with-handlers ([(λ(x) #t) (lambda-s (x) (display x) (newline) #t)])
                        (^up (proceed a))))))))))
  
  (test
   (deploy-fluid (aspect (call test-function) ;; pc
                         handler)                 
                 (test-function 'b))
   #t))


;; Base handler catches aspect exception
(define (test-function-2 x) x)

(let ((raiser 
       (lambda (proceed)   ;; proceed
         (lambda ctx       ;; ctx
           (lambda (a)     ;; val
             (begin
               (printf "Advice. a: ~s - ctx: ~s - proceed: ~s" a ctx proceed)
               (newline)
               (error a)))))))
  
  (test
   (deploy-fluid (aspect (call test-function-2)  ;; pc
                         raiser)                 ;; adv                 
                 (with-handlers ([(λ(x) #t) (^up (lambda-s (x) #t))])
                   (test-function-2 'c)))
   #t))